__all__ = ()

JSON_KEY_BOT_INFO_ID = 'id'
JSON_KEY_BOT_INFO_NAME = 'username'
JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING = 'discriminator'
JSON_KEY_BOT_INFO_AVATAR_BASE64 = 'avatar'
JSON_KEY_BOT_INFO_BANNER_URL = 'bannerUrl'
JSON_KEY_BOT_INFO_PREFIX = 'prefix'
JSON_KEY_BOT_INFO_SHORT_DESCRIPTION = 'shortdesc'
JSON_KEY_BOT_INFO_LONG_DESCRIPTION = 'longdesc'
JSON_KEY_BOT_INFO_TAG_ARRAY = 'tags'
JSON_KEY_BOT_INFO_WEBSITE_URL = 'website'
JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL = 'support'
JSON_KEY_BOT_INFO_GITHUB_URL = 'github'
JSON_KEY_BOT_INFO_OWNER_ID_ARRAY = 'owners'
JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY = 'guilds'
JSON_KEY_BOT_INFO_INVITE_URL = 'invite'
JSON_KEY_BOT_INFO_CERTIFIED_AT = 'date'
JSON_KEY_BOT_INFO_IS_CERTIFIED = 'certifiedBot'
JSON_KEY_BOT_INFO_VANITY_URL = 'vanity'
JSON_KEY_BOT_INFO_UPVOTES = 'points'
JSON_KEY_BOT_INFO_UPVOTES_MONTHLY = 'monthlyPoints'
JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID = 'donatebotguildid'

JSON_KEY_BOT_STATS_GUILD_COUNT = 'serverCount'
JSON_KEY_BOT_STATS_GUILD_COUNT_PER_SHARD_ARRAY = 'shards'
JSON_KEY_BOT_STATS_SHARD_ID = 'shardId'
JSON_KEY_BOT_STATS_SHARD_COUNT = 'shardCount'

JSON_KEY_POST_BOT_STATS_GUILD_COUNT = 'server_count'
JSON_KEY_POST_BOT_STATS_SHARD_ID = 'shard_id'
JSON_KEY_POST_BOT_STATS_SHARD_COUNT = 'shard_count'

JSON_KEY_USER_INFO_ID = 'id'
JSON_KEY_USER_INFO_NAME = 'username'
JSON_KEY_USER_INFO_DISCRIMINATOR_STRING = 'discriminator'
JSON_KEY_USER_INFO_AVATAR_BASE64 = 'avatar'
JSON_KEY_USER_INFO_BIO = 'bio'
JSON_KEY_USER_INFO_BANNER_URL = 'banner'
JSON_KEY_USER_INFO_CONNECTIONS = 'social'
JSON_KEY_USER_INFO_COLOR = 'color'
JSON_KEY_USER_INFO_IS_SUPPORTER = 'supporter'
JSON_KEY_USER_INFO_IS_CERTIFIED_DEVELOPER = 'certifiedDev'
JSON_KEY_USER_INFO_IS_MODERATOR = 'mod'
JSON_KEY_USER_INFO_IS_WEBSITE_MODERATOR = 'webMod'
JSON_KEY_USER_INFO_IS_ADMIN = 'admin'

JSON_KEY_USER_INFO_CONNECTION_YOUTUBE = 'youtube'
JSON_KEY_USER_INFO_CONNECTION_REDDIT = 'reddit'
JSON_KEY_USER_INFO_CONNECTION_TWITTER = 'twitter'
JSON_KEY_USER_INFO_CONNECTION_INSTAGRAM = 'instagram'
JSON_KEY_USER_INFO_CONNECTION_GITHUB = 'github'

JSON_KEY_WEEKEND_STATUS = 'is_weekend'
JSON_KEY_VOTED = 'voted'
JSON_KEY_EXCEPTION_MESSAGE = 'error'

QUERY_KEY_GET_BOTS_LIMIT = 'limit'
QUERY_KEY_GET_BOTS_OFFSET = 'offset'
QUERY_KEY_GET_BOTS_SORT_BY = 'sort'
QUERY_KEY_GET_BOTS_SEARCH_QUERY = 'search'
QUERY_KEY_GET_BOTS_FIELDS = 'fields'

JSON_KEY_BOTS_QUERY_RESULT_RESULTS = 'results'
JSON_KEY_BOTS_QUERY_RESULT_LIMIT = 'limit'
JSON_KEY_BOTS_QUERY_RESULT_OFFSET = 'offset'
JSON_KEY_BOTS_QUERY_RESULT_COUNT = 'count'
JSON_KEY_BOTS_QUERY_RESULT_TOTAL = 'total'

QUERY_KEY_GET_USER_VOTE_USER_ID = 'userId'

RATE_LIMIT_GLOBAL_SIZE = 100
RATE_LIMIT_GLOBAL_RESET_AFTER = 1.0

RATE_LIMIT_BOTS_SIZE = 60
RATE_LIMIT_BOTS_RESET_AFTER = 60.0

RATE_LIMIT_GLOBAL_DEFAULT_DURATION = 3600.0

JSON_KEY_VOTE_BASE_QUERY = 'query'
JSON_KEY_VOTE_BASE_TYPE = 'type'
JSON_KEY_VOTE_BASE_USER_ID = 'user'

JSON_KEY_BOT_VOTE_BOT_ID = 'bot'
JSON_KEY_BOT_VOTE_IS_WEEKEND = 'isWeekend'

JSON_KEY_GUILD_VOTE_GUILD_ID = 'guild'
